@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1999-2002 Free Software Foundation, Inc.
+ * Copyright (C) 1999-2002, 2004-2005 Free Software Foundation, Inc.
  * This file is part of the GNU LIBICONV Library.
  *
  * The GNU LIBICONV Library is free software; you can redistribute it
@@ -14,8 +14,8 @@
  *
  * You should have received a copy of the GNU Library General Public
  * License along with the GNU LIBICONV Library; see the file COPYING.LIB.
- * If not, write to the Free Software Foundation, Inc., 59 Temple Place -
- * Suite 330, Boston, MA 02111-1307, USA.
+ * If not, write to the Free Software Foundation, Inc., 51 Franklin Street,
+ * Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 /* This file defines all the converters. */
@@ -98,6 +98,9 @@ struct conv_struct {
   /* Operation flags */
   int transliterate;
   int discard_ilseq;
+  #ifndef LIBICONV_PLUG
+  struct iconv_hooks hooks;
+  #endif
 };
 
 /*
@@ -139,6 +142,7 @@ struct conv_struct {
 #include "iso8859_8.h"
 #include "iso8859_9.h"
 #include "iso8859_10.h"
+#include "iso8859_11.h"
 #include "iso8859_13.h"
 #include "iso8859_14.h"
 #include "iso8859_15.h"
@@ -176,6 +180,7 @@ struct conv_struct {
 #include "georgian_academy.h"
 #include "georgian_ps.h"
 #include "koi8_t.h"
+#include "pt154.h"
 #include "mulelao.h"
 #include "cp1133.h"
 #include "tis620.h"
@@ -272,7 +277,9 @@ typedef struct {
 #include "euc_jisx0213.h"
 #include "shift_jisx0213.h"
 #include "iso2022_jp3.h"
+#include "big5_2003.h"
 #include "tds565.h"
+#include "atarist.h"
 #include "riscos1.h"
 #endif
 
