@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1999-2003 Free Software Foundation, Inc.
+ * Copyright (C) 1999-2005 Free Software Foundation, Inc.
  * This file is part of the GNU LIBICONV Library.
  *
  * The GNU LIBICONV Library is free software; you can redistribute it
@@ -14,8 +14,8 @@
  *
  * You should have received a copy of the GNU Library General Public
  * License along with the GNU LIBICONV Library; see the file COPYING.LIB.
- * If not, write to the Free Software Foundation, Inc., 59 Temple Place -
- * Suite 330, Boston, MA 02111-1307, USA.
+ * If not, write to the Free Software Foundation, Inc., 51 Franklin Street,
+ * Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #include <iconv.h>
@@ -132,7 +132,7 @@ static struct encoding const all_encodings[] = {
 /*
  * Alias lookup function.
  * Defines
- *   struct alias { const char* name; unsigned int encoding_index; };
+ *   struct alias { int name; unsigned int encoding_index; };
  *   const struct alias * aliases_lookup (const char *str, unsigned int len);
  *   #define MAX_WORD_LENGTH ...
  */
@@ -144,19 +144,21 @@ static struct encoding const all_encodings[] = {
  *   const struct alias * aliases2_lookup (const char *str);
  */
 #if defined(USE_AIX) || defined(USE_OSF1) || defined(USE_DOS) || defined(USE_EXTRA) /* || ... */
-static struct alias sysdep_aliases[] = {
-#ifdef USE_AIX
-#include "aliases_aix.h"
-#endif
-#ifdef USE_OSF1
-#include "aliases_osf1.h"
-#endif
-#ifdef USE_DOS
-#include "aliases_dos.h"
-#endif
-#ifdef USE_EXTRA
-#include "aliases_extra.h"
-#endif
+struct stringpool2_t {
+#define S(tag,name,encoding_index) char stringpool_##tag[sizeof(name)];
+#include "aliases2.h"
+#undef S
+};
+static const struct stringpool2_t stringpool2_contents = {
+#define S(tag,name,encoding_index) name,
+#include "aliases2.h"
+#undef S
+};
+#define stringpool2 ((const char *) &stringpool2_contents)
+static const struct alias sysdep_aliases[] = {
+#define S(tag,name,encoding_index) { (int)(long)&((struct stringpool2_t *)0)->stringpool_##tag, encoding_index },
+#include "aliases2.h"
+#undef S
 };
 #ifdef __GNUC__
 __inline
@@ -164,15 +166,16 @@ __inline
 const struct alias *
 aliases2_lookup (register const char *str)
 {
-  struct alias * ptr;
+  const struct alias * ptr;
   unsigned int count;
   for (ptr = sysdep_aliases, count = sizeof(sysdep_aliases)/sizeof(sysdep_aliases[0]); count > 0; ptr++, count--)
-    if (!strcmp(str,ptr->name))
+    if (!strcmp(str, stringpool2 + ptr->name))
       return ptr;
   return NULL;
 }
 #else
 #define aliases2_lookup(str)  NULL
+#define stringpool2  NULL
 #endif
 
 #if 0
@@ -230,15 +233,20 @@ iconv_t iconv_open (const char* tocode, const char* fromcode)
       if (--count == 0)
         goto invalid;
     }
-    if (bp-buf >= 10 && memcmp(bp-10,"//TRANSLIT",10)==0) {
-      bp -= 10;
-      *bp = '\0';
-      transliterate = 1;
-    }
-    if (bp-buf >= 8 && memcmp(bp-8,"//IGNORE",8)==0) {
-      bp -= 8;
-      *bp = '\0';
-      discard_ilseq = 1;
+    for (;;) {
+      if (bp-buf >= 10 && memcmp(bp-10,"//TRANSLIT",10)==0) {
+        bp -= 10;
+        *bp = '\0';
+        transliterate = 1;
+        continue;
+      }
+      if (bp-buf >= 8 && memcmp(bp-8,"//IGNORE",8)==0) {
+        bp -= 8;
+        *bp = '\0';
+        discard_ilseq = 1;
+        continue;
+      }
+      break;
     }
     if (buf[0] == '\0') {
       tocode = locale_charset();
@@ -301,13 +309,18 @@ iconv_t iconv_open (const char* tocode, const char* fromcode)
       if (--count == 0)
         goto invalid;
     }
-    if (bp-buf >= 10 && memcmp(bp-10,"//TRANSLIT",10)==0) {
-      bp -= 10;
-      *bp = '\0';
-    }
-    if (bp-buf >= 8 && memcmp(bp-8,"//IGNORE",8)==0) {
-      bp -= 8;
-      *bp = '\0';
+    for (;;) {
+      if (bp-buf >= 10 && memcmp(bp-10,"//TRANSLIT",10)==0) {
+        bp -= 10;
+        *bp = '\0';
+        continue;
+      }
+      if (bp-buf >= 8 && memcmp(bp-8,"//IGNORE",8)==0) {
+        bp -= 8;
+        *bp = '\0';
+        continue;
+      }
+      break;
     }
     if (buf[0] == '\0') {
       fromcode = locale_charset();
@@ -401,6 +414,11 @@ iconv_t iconv_open (const char* tocode, const char* fromcode)
   /* Initialize the operation flags. */
   cd->transliterate = transliterate;
   cd->discard_ilseq = discard_ilseq;
+  #ifndef LIBICONV_PLUG
+  cd->hooks.uc_hook = NULL;
+  cd->hooks.wc_hook = NULL;
+  cd->hooks.data = NULL;
+  #endif
   /* Initialize additional fields. */
   if (from_wchar != to_wchar) {
     struct wchar_conv_struct * wcd = (struct wchar_conv_struct *) cd;
@@ -458,16 +476,28 @@ int iconvctl (iconv_t icd, int request, void* argument)
     case ICONV_SET_DISCARD_ILSEQ:
       cd->discard_ilseq = (*(const int *)argument ? 1 : 0);
       return 0;
+    case ICONV_SET_HOOKS:
+      if (argument != NULL) {
+        cd->hooks = *(const struct iconv_hooks *)argument;
+      } else {
+        cd->hooks.uc_hook = NULL;
+        cd->hooks.wc_hook = NULL;
+        cd->hooks.data = NULL;
+      }
+      return 0;
     default:
       errno = EINVAL;
       return -1;
   }
 }
 
+/* An alias after its name has been converted from 'int' to 'const char*'. */
+struct nalias { const char* name; unsigned int encoding_index; };
+
 static int compare_by_index (const void * arg1, const void * arg2)
 {
-  const struct alias * alias1 = (const struct alias *) arg1;
-  const struct alias * alias2 = (const struct alias *) arg2;
+  const struct nalias * alias1 = (const struct nalias *) arg1;
+  const struct nalias * alias2 = (const struct nalias *) arg2;
   return (int)alias1->encoding_index - (int)alias2->encoding_index;
 }
 
@@ -496,7 +526,7 @@ void iconvlist (int (*do_one) (unsigned int namescount,
 #define aliascount2  0
 #endif
 #define aliascount  (aliascount1+aliascount2)
-  struct alias aliasbuf[aliascount];
+  struct nalias aliasbuf[aliascount];
   const char * namesbuf[aliascount];
   size_t num_aliases;
   {
@@ -506,20 +536,26 @@ void iconvlist (int (*do_one) (unsigned int namescount,
     j = 0;
     for (i = 0; i < aliascount1; i++) {
       const struct alias * p = &aliases[i];
-      if (p->name[0] != '\0'
+      if (p->name >= 0
           && p->encoding_index != ei_local_char
-          && p->encoding_index != ei_local_wchar_t)
-        aliasbuf[j++] = *p;
+          && p->encoding_index != ei_local_wchar_t) {
+        aliasbuf[j].name = stringpool + p->name;
+        aliasbuf[j].encoding_index = p->encoding_index;
+        j++;
+      }
     }
 #ifndef aliases2_lookup
-    for (i = 0; i < aliascount2; i++)
-      aliasbuf[j++] = sysdep_aliases[i];
+    for (i = 0; i < aliascount2; i++) {
+      aliasbuf[j].name = stringpool2 + sysdep_aliases[i].name;
+      aliasbuf[j].encoding_index = sysdep_aliases[i].encoding_index;
+      j++;
+    }
 #endif
     num_aliases = j;
   }
   /* Sort by encoding_index. */
   if (num_aliases > 1)
-    qsort(aliasbuf, num_aliases, sizeof(struct alias), compare_by_index);
+    qsort(aliasbuf, num_aliases, sizeof(struct nalias), compare_by_index);
   {
     /* Process all aliases with the same encoding_index together. */
     size_t j;
@@ -542,6 +578,118 @@ void iconvlist (int (*do_one) (unsigned int namescount,
 #undef aliascount1
 }
 
+/*
+ * Table of canonical names of encodings.
+ * Instead of strings, it contains offsets into stringpool and stringpool2.
+ */
+static const unsigned short all_canonical[] = {
+#include "canonical.h"
+#ifdef USE_AIX
+#include "canonical_aix.h"
+#endif
+#ifdef USE_OSF1
+#include "canonical_osf1.h"
+#endif
+#ifdef USE_DOS
+#include "canonical_dos.h"
+#endif
+#ifdef USE_EXTRA
+#include "canonical_extra.h"
+#endif
+#include "canonical_local.h"
+};
+
+const char * iconv_canonicalize (const char * name)
+{
+  const char* code;
+  char buf[MAX_WORD_LENGTH+10+1];
+  const char* cp;
+  char* bp;
+  const struct alias * ap;
+  unsigned int count;
+  unsigned int index;
+  const char* pool;
+
+  /* Before calling aliases_lookup, convert the input string to upper case,
+   * and check whether it's entirely ASCII (we call gperf with option "-7"
+   * to achieve a smaller table) and non-empty. If it's not entirely ASCII,
+   * or if it's too long, it is not a valid encoding name.
+   */
+  for (code = name;;) {
+    /* Search code in the table. */
+    for (cp = code, bp = buf, count = MAX_WORD_LENGTH+10+1; ; cp++, bp++) {
+      unsigned char c = * (unsigned char *) cp;
+      if (c >= 0x80)
+        goto invalid;
+      if (c >= 'a' && c <= 'z')
+        c -= 'a'-'A';
+      *bp = c;
+      if (c == '\0')
+        break;
+      if (--count == 0)
+        goto invalid;
+    }
+    for (;;) {
+      if (bp-buf >= 10 && memcmp(bp-10,"//TRANSLIT",10)==0) {
+        bp -= 10;
+        *bp = '\0';
+        continue;
+      }
+      if (bp-buf >= 8 && memcmp(bp-8,"//IGNORE",8)==0) {
+        bp -= 8;
+        *bp = '\0';
+        continue;
+      }
+      break;
+    }
+    if (buf[0] == '\0') {
+      code = locale_charset();
+      /* Avoid an endless loop that could occur when using an older version
+         of localcharset.c. */
+      if (code[0] == '\0')
+        goto invalid;
+      continue;
+    }
+    pool = stringpool;
+    ap = aliases_lookup(buf,bp-buf);
+    if (ap == NULL) {
+      pool = stringpool2;
+      ap = aliases2_lookup(buf);
+      if (ap == NULL)
+        goto invalid;
+    }
+    if (ap->encoding_index == ei_local_char) {
+      code = locale_charset();
+      /* Avoid an endless loop that could occur when using an older version
+         of localcharset.c. */
+      if (code[0] == '\0')
+        goto invalid;
+      continue;
+    }
+    if (ap->encoding_index == ei_local_wchar_t) {
+#if __STDC_ISO_10646__
+      if (sizeof(wchar_t) == 4) {
+        index = ei_ucs4internal;
+        break;
+      }
+      if (sizeof(wchar_t) == 2) {
+        index = ei_ucs2internal;
+        break;
+      }
+      if (sizeof(wchar_t) == 1) {
+        index = ei_iso8859_1;
+        break;
+      }
+#endif
+    }
+    index = ap->encoding_index;
+    break;
+  }
+  return all_canonical[index] + pool;
+ invalid:
+  return name;
+}
+
 int _libiconv_version = _LIBICONV_VERSION;
 
 #if defined __FreeBSD__ && !defined __gnu_freebsd__
